// derived from https://github.com/EstebanBorai/yaml-front-matter
use serde::de::DeserializeOwned;

#[derive(Debug)]
pub struct Document {
    pub metadata: FrontMatter,
    pub content: String,
}

#[derive(Debug)]
pub struct FrontMatter {
    pub layout: String,
    pub title: String,
    pub date: String,
    pub published: bool,
}

pub struct YamlFrontMatter;
impl YamlFrontMatter {
    pub fn parse<T: DeserializeOwned>(
        markdown: &str,
    ) -> Result<Document, Box<dyn std::error::Error>> {
        let yaml = YamlFrontMatter::extract(markdown)?;
        println!("File front matter metadata (raw): {:?}", yaml.0);
        // println!("File content: {:?}", yaml.1);
        let clean_yaml = YamlFrontMatter::unescape_str(&yaml.0);
        println!("File front matter metadata (clean): {:?}", clean_yaml);
        let metadata = match YamlFrontMatter::from_yaml_str(clean_yaml.as_str()) {
            Ok(m) => m,
            Err(e) => {
                println!("{e}");
                panic!();
            }
        };

        Ok(Document {
            metadata,
            content: yaml.1,
        })
    }

    fn extract(markdown: &str) -> Result<(String, String), Box<dyn std::error::Error>> {
        let mut front_matter = String::default();
        let mut sentinel = false;
        let mut front_matter_lines = 0;
        let lines = markdown.lines();

        for line in lines.clone() {
            front_matter_lines += 1;

            if line.trim() == "---" {
                if sentinel {
                    break;
                }

                sentinel = true;
                continue;
            }

            if sentinel {
                front_matter.push_str(line);
                front_matter.push('\n');
            }
        }

        Ok((
            front_matter,
            lines
                .skip(front_matter_lines)
                .collect::<Vec<&str>>()
                .join("\n"),
        ))
    }

    fn unescape_str(s: &str) -> String {
        s.replace("\\n", "\n")
            .replace("\\\"", "\"")
            .replace("\\\\", "\\")
        // .replace("\\t", "\t")
        // .replace("\\r", "\r")
    }

    fn from_yaml_str(yaml: &str) -> Result<FrontMatter, String> {
        let mut layout = String::new();
        let mut title = String::new();
        let mut date = String::new();
        let mut published = false;

        for line in yaml.lines() {
            let line = line.trim();
            if let Some((key, value)) = line.split_once(':') {
                let key = key.trim();
                let value = value.trim();

                match key {
                    "layout" => layout = value.to_string(),
                    "title" => {
                        // Remove quotes if present
                        title = value.trim_matches('\'').trim_matches('"').to_string();
                    }
                    "date" => date = value.to_string(),
                    "published" => {
                        published = value.parse().map_err(|_| "Invalid boolean for published")?;
                    }
                    _ => {} // Ignore unknown fields
                }
            }
        }

        Ok(FrontMatter {
            layout,
            title,
            date,
            published,
        })
    }
}
